from django.dispatch import Signal

qbd_task_create = Signal(providing_args=[
    "qb_operation",
    "qb_resource",
    "object_id",
    "content_type",
    "realm_id",
    "instance",
])

customer_created = Signal(providing_args=["qbd_model_mixin_obj", "realm_id"])
customer_updated = Signal(providing_args=["qbd_model_mixin_obj", "realm_id"])
invoice_created = Signal(providing_args=["qbd_model_mixin_obj", "realm_id"])
invoice_updated = Signal(providing_args=["qbd_model_mixin_obj", "realm_id"])
realm_authenticated = Signal(providing_args=["realm"])
qbd_first_time_connected = Signal(providing_args=["realm_id"])

from django_quickbooks.signals.customer import *
from django_quickbooks.signals.invoice import *
from django_quickbooks.signals.qbd_task import *



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.utils.decorators import method_decorator
from lxml import etree

from django_quickbooks import get_realm_session_model, get_realm_model, get_qbd_task_model
from django_quickbooks.core.session_manager import BaseSessionManager
from django_quickbooks.decorators import realm_connection
from django_quickbooks.exceptions import QBXMLParseError, QBXMLStatusError, QbException
from django_quickbooks.queue_manager import RabbitMQManager

Realm = get_realm_model()
RealmSession = get_realm_session_model()
QBDTask = get_qbd_task_model()


def _get_realm(ticket):
    return RealmSession.objects.get(id=ticket).realm


class SessionManager(BaseSessionManager, RabbitMQManager):

    def authenticate(self, username, password):
        return Realm.objects.find(username, password)

    def set_ticket(self, realm):
        return str(RealmSession.objects.create(realm=realm).id)

    def in_session(self, realm):
        return RealmSession.objects.is_active(realm)

    @method_decorator(realm_connection())
    def add_new_jobs(self, realm=None):
        queryset = QBDTask.objects.filter(realm=realm).order_by('created_at')
        for qb_task in queryset:
            try:
                self.publish_message(qb_task.get_request(), str(realm.id))
            except QbException as exc:
                logger = logging.getLogger('django.request')
                logger.error(exc.detail)
            except ObjectDoesNotExist:
                pass

        queryset.delete()

    def new_jobs(self, realm):
        return self.get_queue_message_count(str(realm.id))

    def process_response(self, ticket, response, hresult, message):
        from django_quickbooks import get_processors
        realm = _get_realm(ticket)
        processors = get_processors()
        for processor in processors:
            try:
                processed = processor(response, hresult, message).process(realm)
                if processed:
                    break

            except QBXMLParseError as exc:
                logger = logging.getLogger('django.request')
                logger.error(exc.detail)
                return -1
            except QBXMLStatusError as exc:
                logger = logging.getLogger('django.request')
                logger.error(exc.detail)
                return -1

        self._continue_iterative_response(ticket, response)

        jobs_count = self.new_jobs(realm)
        jobs_count = jobs_count if jobs_count <= 98 else 98

        if not jobs_count:
            self.purge_queue(str(realm.id))

        return int(100 / (jobs_count + 1))

    def get_request(self, ticket):
        request = self.get_message(_get_realm(ticket).id)
        if self.is_iterating_request(request):
            self._set_iterative_request(ticket, request)
        return request

    def is_iterating_request(self, request):
        root = etree.fromstring(bytes(str(request), encoding='utf-8'))
        isIterator = root.xpath('boolean(//@iterator)')
        return isIterator

    def clear_ticket(self, ticket):
        RealmSession.objects.close_session(_get_realm(ticket))
        self.delete_queue(ticket)
        return ticket

    def _get_iterative_request(self, ticket):
        return self.get_message(ticket)

    def _set_iterative_request(self, ticket, request):
        self.publish_message(request, ticket, delete_queue=True)

    def _continue_iterative_response(self, ticket, response):
        root = etree.fromstring(bytes(str(response), encoding='utf-8'))
        isIterator = root.xpath('boolean(//@iteratorID)')
        if isIterator:
            reqXML = self._get_iterative_request(ticket)
            reqroot = etree.fromstring(bytes(str(reqXML), encoding='utf-8'))
            iteratorRemainingCount = int(root.xpath('string(//@iteratorRemainingCount)'))
            iteratorID = root.xpath('string(//@iteratorID)')
            if iteratorRemainingCount:
                iteratornode = reqroot.xpath('//*[@iterator]')
                iteratornode[0].set('iterator', 'Continue')
                request_id_path = reqroot.xpath('//@requestID')
                if request_id_path:
                    requestID = int(request_id_path[0])
                    iteratornode[0].set('requestID', str(requestID + 1))
                iteratornode[0].set('iteratorID', iteratorID)
                ntree = etree.ElementTree(reqroot)
                nextreqXML = etree.tostring(ntree, xml_declaration=True, encoding='UTF-8')
                self.publish_message(nextreqXML, _get_realm(ticket).id)



from django.core.exceptions import ObjectDoesNotExist
from django.dispatch import receiver

from django_quickbooks import get_qbd_task_model, QUICKBOOKS_ENUMS, get_realm_model
from django_quickbooks.signals import qbd_task_create

QBDTaskModel = get_qbd_task_model()
RealmModel = get_realm_model()


@receiver(qbd_task_create)
def create_qbd_task(sender, qb_operation, qb_resource, object_id, content_type, realm_id, instance=None, *args, **kwargs):
    try:
        realm = RealmModel.objects.get(id=realm_id)
        data = dict(
            qb_resource=qb_resource,
            object_id=object_id,
            content_type=content_type,
            realm=realm
        )

        if QBDTaskModel.objects \
                .filter(qb_operation=qb_operation, **data) \
                .count():
            return

        if qb_operation == QUICKBOOKS_ENUMS.OPP_MOD and QBDTaskModel.objects \
                .filter(qb_operation=QUICKBOOKS_ENUMS.OPP_ADD, **data) \
                .count():
            return

        QBDTaskModel.objects.create(qb_operation=qb_operation, **data)

    except ObjectDoesNotExist:
        pass



from django.urls import path
from spyne.protocol.soap import Soap11
from spyne.server.django import DjangoView

from django_quickbooks.views import QuickBooksService, Support

urlpatterns = [
    path('quickbooks-desktop/support/', Support.as_view()),
    path('quickbooks-desktop/', DjangoView.as_view(
        services=[QuickBooksService],
        tns='http://developer.intuit.com/',
        in_protocol=Soap11(validator='lxml'),
        out_protocol=Soap11())
         ),
]




from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from django_quickbooks import QUICKBOOKS_ENUMS, qbwc_settings
from django_quickbooks.objects.invoice import Invoice
from django_quickbooks.processors.base import ResponseProcessor, ResponseProcessorMixin

LocalInvoice = qbwc_settings.LOCAL_MODEL_CLASSES['Invoice']


class InvoiceAddResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    local_model_class = LocalInvoice
    obj_class = Invoice

    resource = QUICKBOOKS_ENUMS.RESOURCE_INVOICE
    op_type = QUICKBOOKS_ENUMS.OPP_ADD

    def process(self, realm):
        cont = super().process(realm)
        if not cont:
            return False
        for invoice_ret in list(self._response_body):
            invoice = self.obj_class.from_lxml(invoice_ret)
            local_invoice = None

            if invoice.TxnID:
                local_invoice = self.find_by_id(invoice.Memo)

            if local_invoice:
                self.update(local_invoice, invoice)

        return True

    def find_by_id(self, id):
        try:
            return self.local_model_class.objects.get(id=id)
        except ObjectDoesNotExist:
            return None

    def update(self, local_obj, obj):
        local_obj.qbd_object_id = obj.TxnID
        local_obj.qbd_object_updated_at = timezone.now() + timezone.timedelta(minutes=1)
        local_obj.qbd_object_version = obj.EditSequence
        local_obj.save()


class InvoiceModResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    local_model_class = LocalInvoice
    obj_class = Invoice

    resource = QUICKBOOKS_ENUMS.RESOURCE_INVOICE
    op_type = QUICKBOOKS_ENUMS.OPP_MOD


class InvoiceQueryResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    local_model_class = LocalInvoice
    obj_class = Invoice

    resource = QUICKBOOKS_ENUMS.RESOURCE_INVOICE
    op_type = QUICKBOOKS_ENUMS.OPP_QR



from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from django.utils.decorators import method_decorator
from lxml import etree

from django_quickbooks import QBXML_RESPONSE_STATUS_CODES
from django_quickbooks.decorators import realm_connection
from django_quickbooks.exceptions import QBXMLParseError, QBXMLStatusError


class ResponseProcessor:
    resource = None
    op_type = None
    obj_class = None

    def __init__(self, response, hresult, message):
        self._actual_response_type = None
        self._response_body = None
        self._response = response
        self.hresult = hresult
        self.message = message
        self._process()

    def _process(self):
        if self.hresult:
            raise QBXMLStatusError(self.message)
        qbxml_root = etree.fromstring(self._response)
        if qbxml_root.tag != 'QBXML':
            raise QBXMLParseError('QBXML tag not found')
        if len(qbxml_root) != 1:
            raise QBXMLParseError('QBXML has more or less than 1 child')
        qbxml_msg_rs = qbxml_root[0]
        if qbxml_msg_rs.tag != 'QBXMLMsgsRs':
            raise QBXMLParseError('QBXMLMsgsRs tag not found')
        if len(qbxml_msg_rs) != 1:
            raise QBXMLParseError('QBXMLMsgsRs has more or less than 1 child')
        response_body_root = qbxml_msg_rs[0]
        if 'statusCode' not in response_body_root.attrib:
            raise QBXMLParseError('statusCode is not found in %s' % response_body_root.tag)
        self._actual_response_type = response_body_root.tag
        self._response_body = response_body_root
        self.status_code = response_body_root.attrib['statusCode']
        self.status_message = response_body_root.attrib.get('statusMessage', '')

    def is_valid(self) -> bool:
        return '%s%sRs' % (self.resource, self.op_type) == self._actual_response_type

    @method_decorator(realm_connection())
    def process(self, realm):
        assert self.resource, 'resource attribute is not defined during class definition ' \
                              'of %s' % self.__class__.__name__
        assert self.op_type, 'op_type attribute is not defined during class definition ' \
                             'of %s' % self.__class__.__name__
        if not self.is_valid():
            return False

        if self.status_code != QBXML_RESPONSE_STATUS_CODES.OK:
            raise QBXMLStatusError(self.status_message)

        return True


class ResponseProcessorMixin:
    local_model_class = None

    def update(self, local_obj, obj):
        if local_obj.qbd_object_version != obj.EditSequence:
            local_obj.name = obj.Name
            local_obj.qbd_object_id = obj.ListID
            local_obj.qbd_object_updated_at = timezone.now() + timezone.timedelta(minutes=1)
            local_obj.qbd_object_version = obj.EditSequence
            local_obj.save()

    def find_by_list_id(self, list_id):
        try:
            return self.local_model_class.objects.get(qbd_object_id=list_id)
        except ObjectDoesNotExist:
            return None

    def find_by_name(self, name, field_name='name'):
        try:
            return self.local_model_class.objects.get(**{field_name: name})
        except ObjectDoesNotExist:
            return None

    def create(self, obj):
        customer = self.local_model_class.from_qbd_obj(obj)
        customer.save()



# Generated by Django 2.2.4 on 2020-01-03 08:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_quickbooks', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='realm',
            name='is_active',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='realm',
            name='password',
            field=models.CharField(max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='realm',
            name='schema_name',
            field=models.CharField(max_length=100, unique=True),
        ),
    ]



# Generated by Django 2.2.4 on 2020-01-17 10:01

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('django_quickbooks', '0002_auto_20200103_0801'),
    ]

    operations = [
        migrations.CreateModel(
            name='QBTask',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('qb_operation', models.CharField(max_length=25)),
                ('qb_resource', models.CharField(max_length=50)),
                ('object_id', models.CharField(max_length=50, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('realm', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='qb_tasks', to='django_quickbooks.Realm')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import pytest
from lxml import etree


@pytest.fixture
def sample_customer_xml():
    return """
    <CustomerRet> <!-- optional -->
        <ListID>800004ED-1525972764</ListID> <!-- required -->
        <TimeCreated>2018-03-24T00:31:04+05:00</TimeCreated> <!-- required -->
        <TimeModified>2018-03-24T00:31:04+05:00</TimeModified> <!-- required -->
        <EditSequence>1525972764</EditSequence> <!-- required -->
        <Name>Amazon</Name> <!-- required -->
        <FullName>Amazon</FullName> <!-- required -->
        <IsActive>true</IsActive> <!-- optional -->
        <CompanyName>Amazon</CompanyName> <!-- optional -->
        <BillAddress> <!-- optional -->
            <Addr1>2305 Litton Ln</Addr1> <!-- optional -->
            <Addr2></Addr2> <!-- optional -->
            <Addr3></Addr3> <!-- optional -->
            <Addr4></Addr4> <!-- optional -->
            <Addr5></Addr5> <!-- optional -->
            <City>Hebron</City> <!-- optional -->
            <State>Kentucky</State> <!-- optional -->
            <PostalCode>41048</PostalCode> <!-- optional -->
            <Country>United States</Country> <!-- optional -->
            <Note>Nice address</Note> <!-- optional -->
        </BillAddress>
        <ShipAddress> <!-- optional -->
            <Addr1>2305 Litton Ln</Addr1> <!-- optional -->
            <Addr2></Addr2> <!-- optional -->
            <Addr3></Addr3> <!-- optional -->
            <Addr4></Addr4> <!-- optional -->
            <Addr5></Addr5> <!-- optional -->
            <City>Hebron</City> <!-- optional -->
            <State>Kentucky</State> <!-- optional -->
            <PostalCode>41048</PostalCode> <!-- optional -->
            <Country>United States</Country> <!-- optional -->
            <Note>Nice address</Note> <!-- optional -->
        </ShipAddress>
        <Phone>998909090909</Phone> <!-- optional -->
        <AltPhone>998909090910</AltPhone> <!-- optional -->
        <Fax>998909090911</Fax> <!-- optional -->
        <Email>info@amazon.com</Email> <!-- optional -->
        <Contact>Someone from Amazon</Contact> <!-- optional -->
        <AltContact>Some other one from Amazon</AltContact> <!-- optional -->
    </CustomerRet>
    """


@pytest.fixture
def sample_customer_data():
    return dict(
        ListID='800004ED-1525972764',
        TimeCreated='2018-03-24T00:31:04+05:00',
        TimeModified='2018-03-24T00:31:04+05:00',
        EditSequence='1525972764',
        Name='Amazon',
        FullName='Amazon',
        IsActive=True,
        CompanyName='Amazon',
        Phone='998909090909',
        AltPhone='998909090910',
        Fax='998909090911',
        Email='info@amazon.com',
        Contact='Someone from Amazon',
        AltContact='Some other one from Amazon',
    )


@pytest.fixture
def sample_address_xml():
    return """
    <Address> <!-- optional -->
        <Addr1>2305 Litton Ln</Addr1> <!-- optional -->
        <Addr2></Addr2> <!-- optional -->
        <Addr3></Addr3> <!-- optional -->
        <Addr4></Addr4> <!-- optional -->
        <Addr5></Addr5> <!-- optional -->
        <City>Hebron</City> <!-- optional -->
        <State>Kentucky</State> <!-- optional -->
        <PostalCode>41048</PostalCode> <!-- optional -->
        <Country>United States</Country> <!-- optional -->
        <Note>Nice address</Note> <!-- optional -->
    </Address>
    """


@pytest.fixture
def sample_address_data():
    return dict(
        Addr1='2305 Litton Ln',
        City='Hebron',
        State='Kentucky',
        PostalCode='41048',
        Country='United States',
        Note='Nice address'
    )

@pytest.fixture
def sample_invoice_xml():
    return """
    <InvoiceRet> <!-- optional -->
        <TxnID>800004ED-1525972764</TxnID> <!-- required -->
        <TimeCreated>2018-03-24T00:31:04+05:00</TimeCreated> <!-- required -->
        <TimeModified>2018-03-24T00:31:04+05:00</TimeModified> <!-- required -->
        <EditSequence>1525972764</EditSequence> <!-- required -->
        <CustomerRef> <!-- required -->
            <ListID>800004ED-1525972764</ListID> <!-- optional -->
            <FullName>Amazon</FullName> <!-- optional -->
        </CustomerRef>
        <TxnDate>2018-03-24T00:31:04+05:00</TxnDate> <!-- required -->
        <BillAddress> <!-- optional -->
            <Addr1>2305 Litton Ln</Addr1> <!-- optional -->
            <Addr2></Addr2> <!-- optional -->
            <Addr3></Addr3> <!-- optional -->
            <Addr4></Addr4> <!-- optional -->
            <Addr5></Addr5> <!-- optional -->
            <City>Hebron</City> <!-- optional -->
            <State>Kentucky</State> <!-- optional -->
            <PostalCode>41048</PostalCode> <!-- optional -->
            <Country>United States</Country> <!-- optional -->
            <Note>Nice address</Note> <!-- optional -->
        </BillAddress>
        <ShipAddress> <!-- optional -->
            <Addr1>2305 Litton Ln</Addr1> <!-- optional -->
            <Addr2></Addr2> <!-- optional -->
            <Addr3></Addr3> <!-- optional -->
            <Addr4></Addr4> <!-- optional -->
            <Addr5></Addr5> <!-- optional -->
            <City>Hebron</City> <!-- optional -->
            <State>Kentucky</State> <!-- optional -->
            <PostalCode>41048</PostalCode> <!-- optional -->
            <Country>United States</Country> <!-- optional -->
            <Note>Nice address</Note> <!-- optional -->
        </ShipAddress>
        <IsPending>true</IsPending> <!-- optional -->
        <DueDate>2018-03-24T00:31:04+05:00</DueDate> <!-- optional -->
        <Memo>a2e8929f-6f03-4bc2-ad27-ea6aa2cc11cd</Memo> <!-- optional -->
        <InvoiceLineRet> <!-- optional -->
            <TxnLineID>800004ED-1525972764</TxnLineID> <!-- required -->
            <ItemRef> <!-- optional -->
                <ListID>800004ED-1525972764</ListID> <!-- optional -->
                <FullName>Line Haul</FullName> <!-- optional -->
            </ItemRef>
            <Desc>First Line of Invoice</Desc> <!-- optional -->
            <Quantity>1.0</Quantity> <!-- optional -->
            <Rate>500.5</Rate> <!-- optional -->
        </InvoiceLineRet>
        <InvoiceLineRet> <!-- optional -->
            <TxnLineID>800004ED-1525972765</TxnLineID> <!-- required -->
            <ItemRef> <!-- optional -->
                <ListID>800004ED-1525972765</ListID> <!-- optional -->
                <FullName>Fuel Surcharge</FullName> <!-- optional -->
            </ItemRef>
            <Desc>Second Line of Invoice</Desc> <!-- optional -->
            <Quantity>1.0</Quantity> <!-- optional -->
            <Rate>100.5</Rate> <!-- optional -->
        </InvoiceLineRet>
    </InvoiceRet>
    """


@pytest.fixture
def sample_invoice_customer_data():
    return dict(
        ListID='800004ED-1525972764',
        FullName='Amazon',
    )


@pytest.fixture
def sample_invoice_item_service1_data():
    return dict(
        ListID='800004ED-1525972764',
        FullName='Line Haul',
    )


@pytest.fixture
def sample_invoice_item_service2_data():
    return dict(
        ListID='800004ED-1525972765',
        FullName='Fuel Surcharge',
    )


@pytest.fixture
def sample_invoice_line1_data():
    return dict(
        TxnLineID='800004ED-1525972764',
        Desc='First Line of Invoice',
        Quantity=1.0,
        Rate=500.5,
    )


@pytest.fixture
def sample_invoice_line2_data():
    return dict(
        TxnLineID='800004ED-1525972765',
        Desc='Second Line of Invoice',
        Quantity=1.0,
        Rate=100.5,
    )


@pytest.fixture
def sample_invoice_data():
    return dict(
        TxnID='800004ED-1525972764',
        TimeCreated='2018-03-24T00:31:04+05:00',
        TimeModified='2018-03-24T00:31:04+05:00',
        EditSequence='1525972764',
        TxnDate='2018-03-24T00:31:04+05:00',
        IsPending=True,
        DueDate='2018-03-24T00:31:04+05:00',
        Memo='a2e8929f-6f03-4bc2-ad27-ea6aa2cc11cd',
    )


@pytest.fixture()
def xml_compare_fn():
    def remove_whitespaces(xml):
        tree = etree.fromstring(xml)
        root = tree.getroottree()

        for elem in root.iter('*'):
            if elem.text is not None:
                elem.text = elem.text.strip()
            if elem.tail is not None:
                elem.tail = elem.tail.strip()
        return root

    def compare(xml1, xml2):
        tree1 = remove_whitespaces(xml1)
        tree2 = remove_whitespaces(xml2)

        assert etree.tostring(tree1, encoding="utf-8", xml_declaration=True) == etree.tostring(tree2, encoding="utf-8", xml_declaration=True)

    return compare



from copy import deepcopy

import pytest

from django_quickbooks.objects import BillAddress, ShipAddress, Customer
from django_quickbooks.services.customer import CustomerService


@pytest.fixture
def customer_service():
    return CustomerService()


@pytest.fixture
def sample_customer_add(
        sample_customer_data,
        sample_address_data,
):
    customer_data = deepcopy(sample_customer_data)
    customer_data.pop('ListID')
    customer_data.pop('TimeCreated')
    customer_data.pop('TimeModified')
    customer_data.pop('EditSequence')

    return Customer(
        **customer_data,
        BillAddress=BillAddress(**sample_address_data),
        ShipAddress=ShipAddress(**sample_address_data),
    )


@pytest.fixture
def sample_customer_add_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <CustomerAddRq>
            <CustomerAdd>
                <Name>Amazon</Name>
                <FullName>Amazon</FullName>
                <IsActive>true</IsActive>
                <CompanyName>Amazon</CompanyName>
                <BillAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </BillAddress>
                <ShipAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </ShipAddress>
                <Phone>998909090909</Phone>
                <AltPhone>998909090910</AltPhone>
                <Fax>998909090911</Fax>
                <Email>info@amazon.com</Email>
                <Contact>Someone from Amazon</Contact>
                <AltContact>Some other one from Amazon</AltContact>
            </CustomerAdd>
        </CustomerAddRq>
    </QBXMLMsgsRq>
</QBXML>"""


def test_customer_service_add(
        customer_service,
        sample_customer_add,
        sample_customer_add_xml,
        xml_compare_fn
):
    checking_xml = customer_service.add(sample_customer_add)
    xml_compare_fn(checking_xml, sample_customer_add_xml)


@pytest.fixture
def sample_customer_mod(
        sample_customer_add,
        sample_customer_data
):
    customer = deepcopy(sample_customer_add)
    customer.ListID = sample_customer_data['ListID']
    customer.EditSequence = sample_customer_data['EditSequence']

    return customer


@pytest.fixture
def sample_customer_mod_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <CustomerModRq>
            <CustomerMod>
                <ListID>800004ED-1525972764</ListID>
                <EditSequence>1525972764</EditSequence>
                <Name>Amazon</Name>
                <FullName>Amazon</FullName>
                <IsActive>true</IsActive>
                <CompanyName>Amazon</CompanyName>
                <BillAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </BillAddress>
                <ShipAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </ShipAddress>
                <Phone>998909090909</Phone>
                <AltPhone>998909090910</AltPhone>
                <Fax>998909090911</Fax>
                <Email>info@amazon.com</Email>
                <Contact>Someone from Amazon</Contact>
                <AltContact>Some other one from Amazon</AltContact>
            </CustomerMod>
        </CustomerModRq>
    </QBXMLMsgsRq>
</QBXML>"""


def test_customer_service_update(
        customer_service,
        sample_customer_mod,
        sample_customer_mod_xml,
        xml_compare_fn
):
    checking_xml = customer_service.update(sample_customer_mod)
    xml_compare_fn(checking_xml, sample_customer_mod_xml)


@pytest.fixture
def sample_customer_query_all_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <CustomerQueryRq metaData="NoMetaData" iterator="Start">
            <MaxReturned>100</MaxReturned>
        </CustomerQueryRq>
    </QBXMLMsgsRq>
</QBXML>"""


def test_customer_service_query_all(
        customer_service,
        sample_customer_query_all_xml,
        xml_compare_fn
):
    checking_xml = customer_service.all()
    xml_compare_fn(checking_xml, sample_customer_query_all_xml)


@pytest.fixture
def sample_customer_query_id_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <CustomerQueryRq metaData="NoMetaData">
            <ListID>800004ED-1525972764</ListID>
        </CustomerQueryRq>
    </QBXMLMsgsRq>
</QBXML>"""


def test_customer_service_query_id(customer_service, sample_customer_query_id_xml, xml_compare_fn):
    checking_xml = customer_service.find_by_id('800004ED-1525972764')
    xml_compare_fn(checking_xml, sample_customer_query_id_xml)


@pytest.fixture
def sample_customer_query_full_name_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <CustomerQueryRq metaData="NoMetaData">
            <FullName>Amazon</FullName>
        </CustomerQueryRq>
    </QBXMLMsgsRq>
</QBXML>"""


def test_customer_service_query_full_name(customer_service, sample_customer_query_full_name_xml, xml_compare_fn):
    checking_xml = customer_service.find_by_full_name('Amazon')
    xml_compare_fn(checking_xml, sample_customer_query_full_name_xml)



from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver

from django_quickbooks import get_realm_model, QUICKBOOKS_ENUMS
from django_quickbooks.signals import customer_created, customer_updated, qbd_task_create

RealmModel = get_realm_model()


@receiver(customer_created)
def create_qbd_customer(sender, qbd_model_mixin_obj, realm_id, *args, **kwargs):
    qbd_task_create.send(
        sender=qbd_model_mixin_obj.__class__,
        qb_operation=QUICKBOOKS_ENUMS.OPP_ADD,
        qb_resource=QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER,
        object_id=qbd_model_mixin_obj.id,
        content_type=ContentType.objects.get_for_model(qbd_model_mixin_obj),
        realm_id=realm_id,
    )


@receiver(customer_updated)
def update_qbd_customer(sender, qbd_model_mixin_obj, realm_id, *args, **kwargs):
    qbd_task_create.send(
        sender=qbd_model_mixin_obj.__class__,
        qb_operation=QUICKBOOKS_ENUMS.OPP_MOD,
        qb_resource=QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER,
        object_id=qbd_model_mixin_obj.id,
        content_type=ContentType.objects.get_for_model(qbd_model_mixin_obj),
        realm_id=realm_id,
    )



from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone


class RealmQuerySet(models.QuerySet):
    def find(self, username, password):
        try:
            realm = self.get(id=username)
            if realm.check_password(password):
                return realm
        except ObjectDoesNotExist:
            pass
        return None


class RealmSessionQuerySet(models.QuerySet):
    def is_active(self, realm):
        try:
            return self.get(realm=realm, ended_at__isnull=True) is not None
        except ObjectDoesNotExist:
            return False

    def close_session(self, realm):
        self.filter(realm=realm, ended_at__isnull=True).update(ended_at=timezone.now())


class QBDTaskQuerySet(models.QuerySet):
    pass



# Generated by Django 2.2.4 on 2019-12-24 14:43

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Realm',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=155)),
                ('password', models.CharField(max_length=128)),
                ('schema_name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RealmSession',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid1, editable=False, primary_key=True, serialize=False, verbose_name='Ticket for QBWC session')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('ended_at', models.DateTimeField(null=True)),
                ('realm', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to='django_quickbooks.Realm')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import logging

import pika
from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.adapters.blocking_connection import BlockingChannel

from django_quickbooks.core.queue_manager import QueueManager
from django_quickbooks.settings import qbwc_settings


class RabbitMQManager(QueueManager):
    def __init__(self):
        self._connection = None
        self._input_channel = None
        self._output_channel = None

    def _get_channel(self, receive=True) -> BlockingChannel:
        if receive:
            if hasattr(self, '_input_channel') and self._input_channel and self._input_channel.is_open:
                return self._input_channel

            self._input_channel = self._get_connection().channel()
            return self._input_channel
        else:
            if hasattr(self, '_output_channel') and self._output_channel and self._output_channel.is_open:
                return self._output_channel

            self._output_channel = self._get_connection().channel()
            return self._output_channel

    def _get_connection(self) -> BlockingConnection:
        if hasattr(self, '_connection') and self._connection and self._connection.is_open:
            return self._connection
        self._connection = BlockingConnection(
            ConnectionParameters(
                host=qbwc_settings.RABBITMQ_DEFAULT_HOST,
                virtual_host=qbwc_settings.RABBITMQ_DEFAULT_VHOST,
                credentials=PlainCredentials(
                    username=qbwc_settings.RABBITMQ_DEFAULT_USER,
                    password=qbwc_settings.RABBITMQ_DEFAULT_PASS
                ),
                heartbeat=600,
                blocked_connection_timeout=300,
                connection_attempts=1000,
            )
        )

        return self._connection

    def get_message(self, queue_name, ack=True):
        channel = self._get_channel()
        method_frame, header_frame, body = channel.basic_get(str(queue_name))
        if method_frame:
            if ack:
                self._get_channel().basic_ack(method_frame.delivery_tag)
            return body if isinstance(body, str) else body.decode('UTF-8') if isinstance(body, bytes) else body
        else:
            logging.error("No request in the queue, error.")
            return ""

    def publish_message(self, msg, queue_name, delete_queue=False):
        queue_name = str(queue_name)
        channel = self._get_channel(receive=False)
        if delete_queue:
            channel.queue_delete(queue_name)
        channel.queue_declare(queue_name, arguments={'x-queue-mode': 'lazy', 'x-max-length': 1000})
        channel.basic_publish(exchange='', routing_key=queue_name, body=msg, properties=pika.BasicProperties(
            delivery_mode=2  # persistent
        ))

    def get_queue_message_count(self, queue_name):
        return self._get_channel().queue_declare(
            queue=queue_name,
            arguments={'x-queue-mode': 'lazy', 'x-max-length': 1000}
        ).method.message_count

    def delete_queue(self, queue_name):
        self._get_channel().queue_delete(queue=queue_name)

    def purge_queue(self, queue_name):
        self._get_channel().queue_purge(queue=queue_name)



from lxml import etree
import pytest

from django_quickbooks.exceptions import ValidationError
from django_quickbooks.objects import BillAddress, ShipAddress


def test_create_bill_address_with_validation_error():
    with pytest.raises(ValidationError):
        BillAddress(
            Addr1=12312,
            City=1234,
            State=1234,
            PostalCode=1234,
            Country=1234,
        )


def test_create_bill_address(sample_address_data):
    address = BillAddress(**sample_address_data)

    assert address.Addr1 == sample_address_data['Addr1']
    assert address.City == sample_address_data['City']
    assert address.State == sample_address_data['State']
    assert address.PostalCode == sample_address_data['PostalCode']
    assert address.Country == sample_address_data['Country']
    assert address.Note == sample_address_data['Note']


def test_create_bill_address_from_xml(sample_address_xml, sample_address_data):
    root_lxml = etree.fromstring(sample_address_xml)
    bill_address = BillAddress.from_lxml(root_lxml)
    assert bill_address.Addr1 == sample_address_data['Addr1']
    assert bill_address.City == sample_address_data['City']
    assert bill_address.State == sample_address_data['State']
    assert bill_address.PostalCode == sample_address_data['PostalCode']
    assert bill_address.Country == sample_address_data['Country']
    assert bill_address.Note == sample_address_data['Note']



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.objects.invoice import ItemService
from django_quickbooks.processors import ResponseProcessor


class ItemServiceQueryResponseProcessor(ResponseProcessor):
    resource = QUICKBOOKS_ENUMS.RESOURCE_ITEM_SERVICE
    op_type = QUICKBOOKS_ENUMS.OPP_QR
    obj_class = ItemService

    def process(self, realm):
        cont = super().process(realm)
        if not cont:
            return False

        return True



from django.db import connection

from django_quickbooks.settings import qbwc_settings


def realm_connection():
    return qbwc_settings.REALM_CONNECTION_DECORATOR


def base_realm_connection(func):
    def connect(realm, *args, **kwargs):
        return func(realm, *args, **kwargs)

    return connect


def base_realm_tenant_connection(func):
    def connect(realm, *args, **kwargs):
        if hasattr(realm, 'schema_name') and hasattr(connection, 'set_schema'):
            try:
                from django_tenants.utils import get_tenant_model
                connection.set_tenant(get_tenant_model().objects.get(schema_name=realm.schema_name))
            except ImportError:
                raise ModuleNotFoundError(
                    'django-tenants package is not installed: pip install django-quickbooks[tenant]'
                )
        return func(realm, *args, **kwargs)

    return connect



from abc import abstractmethod


class QueueManager:

    @abstractmethod
    def get_message(self, queue_name, ack=True):
        pass

    @abstractmethod
    def publish_message(self, msg, queue_name, delete_queue=False):
        pass

    @abstractmethod
    def get_queue_message_count(self, queue_name):
        pass

    @abstractmethod
    def delete_queue(self, queue_name):
        pass

    @abstractmethod
    def purge_queue(self, queue_name):
        pass



from importlib import import_module

from django.conf import settings

DEFAULTS = {
    'UPDATE_PAUSE_SECONDS': 35,
    'MINIMUM_UPDATE_SECONDS': 15,
    'MINIMUM_RUN_EVERY_NSECONDS': 30,
    'MINIMUM_RUN_EVERY_NMINUTES': 15,

    'SESSION_MANAGER_CLASS': 'django_quickbooks.session_manager.SessionManager',

    'REALM_MODEL_CLASS': 'django_quickbooks.models.Realm',
    'REALM_SESSION_MODEL_CLASS': 'django_quickbooks.models.RealmSession',
    'QBD_TASK_MODEL_CLASS': 'django_quickbooks.models.QBDTask',
    
    'REALM_CONNECTION_DECORATOR': 'django_quickbooks.decorators.base_realm_connection',

    'RESPONSE_PROCESSORS': (
        'django_quickbooks.processors.CustomerQueryResponseProcessor',
        'django_quickbooks.processors.CustomerModResponseProcessor',
        'django_quickbooks.processors.CustomerAddResponseProcessor',
        'django_quickbooks.processors.InvoiceQueryResponseProcessor',
        'django_quickbooks.processors.InvoiceAddResponseProcessor',
        'django_quickbooks.processors.InvoiceModResponseProcessor',
        'django_quickbooks.processors.ItemServiceQueryResponseProcessor',
    ),

    'RABBITMQ_DEFAULT_HOST': 'localhost',
    'RABBITMQ_DEFAULT_USER': 'quickbooks',
    'RABBITMQ_DEFAULT_PASS': 'quickbooks',
    'RABBITMQ_DEFAULT_VHOST': 'quickbooks',

    'APP_URL': 'http://localhost:8000/quickbooks-desktop/',
    'APP_SUPPORT': 'http://localhost:8000/quickbooks-desktop/support/',
    'APP_ID': '',
    'APP_NAME': 'Some App',
    'APP_DESCRIPTION': 'Some App Description',
    'QB_TYPE': 'QBFS',
    'OWNER_ID': '{1ee58da6-3051-11ea-b499-9cda3ea7afc1}',

    'LOCAL_MODEL_CLASSES': {
        'Invoice': '',
        'Customer': '',
    }
}

IMPORT_STRINGS = (
    'RESPONSE_PROCESSORS',
    'SESSION_MANAGER_CLASS',
    'REALM_MODEL_CLASS',
    'REALM_SESSION_MODEL_CLASS',
    'QBD_TASK_MODEL_CLASS',
    'REALM_CONNECTION_DECORATOR',
)


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        module_path, class_name = val.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '%s' for Quickbooks Web Connector setting '%s'. %s: %s." \
              % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class QBWCSettings(object):
    """
    A settings object, that allows Quickbooks Web Connector settings to be accessed as properties.
    """

    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'QBWC_SETTINGS', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Quickbooks Web Connector setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        if attr == 'LOCAL_MODEL_CLASSES':
            for key, value in val.items():
                val[key] = import_from_string(value, value)

        setattr(self, attr, val)
        return val


qbwc_settings = QBWCSettings(None, DEFAULTS, IMPORT_STRINGS)



import pytest
from lxml import etree

from django_quickbooks.exceptions import ValidationError
from django_quickbooks.objects import Customer, BillAddress, ShipAddress


def test_create_customer_validation_with_error():
    with pytest.raises(ValidationError):
        Customer(
            ListID=12312,
            EditSequence='asdf',
            Name=1234,
            FullName=1234,
            IsActive='True',
            BillAddress='Somewhere near Amazon',
            CompanyName=1234,
            Phone=1234,
            AltPhone=1234,
            Fax=1234,
            Email=1234,
            Contact=1234,
            AltContact=1234
        )


def test_create_customer(sample_customer_data, sample_address_data):
    customer = Customer(
        **sample_customer_data,
        BillAddress=BillAddress(**sample_address_data),
        ShipAddress=ShipAddress(**sample_address_data),
    )

    assert customer.Name == sample_customer_data['Name']
    assert customer.IsActive == sample_customer_data['IsActive']
    assert customer.ListID == sample_customer_data['ListID']
    assert customer.EditSequence == sample_customer_data['EditSequence']
    assert customer.FullName == sample_customer_data['FullName']
    assert customer.CompanyName == sample_customer_data['CompanyName']
    assert customer.Phone == sample_customer_data['Phone']
    assert customer.AltPhone == sample_customer_data['AltPhone']
    assert customer.Fax == sample_customer_data['Fax']
    assert customer.Email == sample_customer_data['Email']
    assert customer.Contact == sample_customer_data['Contact']
    assert customer.AltContact == sample_customer_data['AltContact']

    assert isinstance(customer.BillAddress, BillAddress)
    assert isinstance(customer.ShipAddress, ShipAddress)


def test_create_customer_from_xml(sample_customer_xml, sample_customer_data, sample_address_data):
    root_lxml = etree.fromstring(sample_customer_xml)
    customer = Customer.from_lxml(root_lxml)
    assert customer.ListID == sample_customer_data['ListID']
    assert customer.TimeCreated == sample_customer_data['TimeCreated']
    assert customer.TimeModified == sample_customer_data['TimeModified']
    assert customer.EditSequence == sample_customer_data['EditSequence']
    assert customer.Name == sample_customer_data['Name']
    assert customer.FullName == sample_customer_data['FullName']
    assert customer.IsActive == sample_customer_data['IsActive']
    assert customer.CompanyName == sample_customer_data['CompanyName']
    assert customer.Phone == sample_customer_data['Phone']
    assert customer.AltPhone == sample_customer_data['AltPhone']
    assert customer.Fax == sample_customer_data['Fax']
    assert customer.Email == sample_customer_data['Email']
    assert customer.Contact == sample_customer_data['Contact']
    assert customer.AltContact == sample_customer_data['AltContact']

    assert customer.BillAddress == BillAddress(**sample_address_data)
    assert customer.ShipAddress == ShipAddress(**sample_address_data)



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.objects.invoice import Txn
from django_quickbooks.services.base import Service


class InvoiceService(Service):
    ref_fields = ['Customer', 'Item']
    add_fields = ['InvoiceLine']
    complex_fields = ['BillAddress']
    mod_fields = ['InvoiceLine']

    def add(self, object):
        return self._add(QUICKBOOKS_ENUMS.RESOURCE_INVOICE, object)

    def update(self, object):
        return self._update(QUICKBOOKS_ENUMS.RESOURCE_INVOICE, object)

    def all(self):
        return self._all(QUICKBOOKS_ENUMS.RESOURCE_INVOICE)

    def void(self, object):
        return self._void(
            QUICKBOOKS_ENUMS.RESOURCE_TXN, Txn(TxnID=object.TxnID, TxnType=QUICKBOOKS_ENUMS.RESOURCE_INVOICE)
        )

    def delete(self, object):
        return self._delete(
            QUICKBOOKS_ENUMS.RESOURCE_TXN, Txn(TxnID=object.TxnID, TxnType=QUICKBOOKS_ENUMS.RESOURCE_INVOICE)
        )

    def find_by_id(self, id):
        return self._find_by_id(QUICKBOOKS_ENUMS.RESOURCE_INVOICE, id, field_name='TxnID')



from abc import abstractmethod

from django_quickbooks import get_realm_session_model, get_realm_model, get_qbd_task_model

Realm = get_realm_model()
RealmSession = get_realm_session_model()
QBDTask = get_qbd_task_model()


class BaseSessionManager:

    @abstractmethod
    def authenticate(self, username: str, password: str) -> Realm:
        pass

    @abstractmethod
    def set_ticket(self, realm: Realm) -> str:
        pass

    @abstractmethod
    def in_session(self, realm: Realm) -> bool:
        pass

    @abstractmethod
    def add_new_jobs(self, realm: Realm) -> None:
        pass

    @abstractmethod
    def new_jobs(self, realm: Realm) -> int:
        pass

    @abstractmethod
    def process_response(self, ticket: str, response: str, hresult: str, message: str) -> int:
        pass

    @abstractmethod
    def get_request(self, ticket: str) -> str:
        pass

    @abstractmethod
    def clear_ticket(self, ticket: str) -> str:
        pass



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.services.base import Service


class CustomerService(Service):
    complex_fields = ['BillAddress', 'ShipAddress']

    def add(self, object):
        return self._add(QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER, object)

    def update(self, object):
        return self._update(QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER, object)

    def all(self):
        return self._all(QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER)

    def find_by_id(self, id):
        return self._find_by_id(QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER, id)

    def find_by_full_name(self, full_name):
        return self._find_by_full_name(QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER, full_name)



from django_quickbooks import QUICKBOOKS_ENUMS, qbwc_settings
from django_quickbooks.objects.customer import Customer
from django_quickbooks.processors.base import ResponseProcessor, ResponseProcessorMixin

LocalCustomer = qbwc_settings.LOCAL_MODEL_CLASSES['Customer']


class CustomerQueryResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    resource = QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER
    op_type = QUICKBOOKS_ENUMS.OPP_QR
    local_model_class = LocalCustomer
    obj_class = Customer

    def process(self, realm):
        cont = super().process(realm)
        if not cont:
            return False
        for customer_ret in list(self._response_body):
            customer = self.obj_class.from_lxml(customer_ret)
            local_customer = None
            if customer.ListID:
                local_customer = self.find_by_list_id(customer.ListID)
            if not local_customer and customer.Name:
                local_customer = self.find_by_name(customer.Name)

            if local_customer:
                self.update(local_customer, customer)
            else:
                self.create(customer)
        return True


class CustomerAddResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    resource = QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER
    op_type = QUICKBOOKS_ENUMS.OPP_ADD
    local_model_class = LocalCustomer
    obj_class = Customer

    def process(self, realm):
        cont = super().process(realm)
        if not cont:
            return False
        for customer_ret in list(self._response_body):
            customer = self.obj_class.from_lxml(customer_ret)
            local_customer = None
            if customer.Name:
                local_customer = self.find_by_name(customer.Name)

            if local_customer:
                self.update(local_customer, customer)
        return True


class CustomerModResponseProcessor(ResponseProcessor, ResponseProcessorMixin):
    resource = QUICKBOOKS_ENUMS.RESOURCE_CUSTOMER
    op_type = QUICKBOOKS_ENUMS.OPP_MOD
    local_model_class = LocalCustomer
    obj_class = Customer

    def process(self, realm):
        cont = super().process(realm)
        if not cont:
            return False
        for customer_ret in list(self._response_body):
            customer = self.obj_class.from_lxml(customer_ret)
            local_customer = None
            if customer.ListID:
                local_customer = self.find_by_list_id(customer.ListID)
            elif not local_customer and customer.Name:
                local_customer = self.find_by_name(customer.Name)

            if local_customer:
                self.update(local_customer, customer)
        return True



from django.contrib import admin

from django_quickbooks import get_realm_model, get_realm_session_model

Realm = get_realm_model()
RealmSession = get_realm_session_model()


class RealmAdmin(admin.ModelAdmin):
    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        if not change:
            obj.set_password(obj.password)
            obj.save()


admin.site.register(Realm, RealmAdmin)
admin.site.register(RealmSession)



from django.core.exceptions import ImproperlyConfigured

from django_quickbooks.settings import qbwc_settings
from django_quickbooks.utils import import_callable

HIGHEST_SUPPORTING_QBWC_VERSION = '2.2.0.34'
default_app_config = 'django_quickbooks.apps.DjangoQuickbooksConfig'


def get_session_manager():
    return qbwc_settings.SESSION_MANAGER_CLASS()


class QBWC_CODES:
    NONE = 'none'  # no work to do for the service
    BUSY = 'busy'  # service is busy with other task
    NVU = 'nvu'  # invalid user is sent to the service
    CC = ''  # indicates current company to use for the web connector to proceed further
    CONN_CLS_OK = 'ok'  # indicates web connector and web service _set_connection closed successfully
    CONN_CLS_ERR = 'done'  # indicates web connector failed connecting to web service and finished its job
    INTR_DONE = 'done'  # indicates web connector finished interactive session with web service
    UNEXP_ERR = 'unexpected error'  # unexpected error received from web connector
    CV = ''  # no update needed for web connector to update its version, it can proceed further


class QUICKBOOKS_ENUMS:
    VERSION_13 = 13
    OPP_ADD = 'Add'
    OPP_MOD = 'Mod'
    OPP_QR = 'Query'
    OPP_DEL = 'Del'
    OPP_VOID = 'Void'

    OPERATIONS = (
        (OPP_ADD, 'Add'),
        (OPP_MOD, 'Modify'),
        (OPP_QR, 'Query'),
        (OPP_DEL, 'Delete'),
    )

    OBJ_REF = 'Ref'

    RESOURCE_CUSTOMER = 'Customer'
    RESOURCE_VENDOR = 'Vendor'
    RESOURCE_INVOICE = 'Invoice'
    RESOURCE_BILL = 'Bill'
    RESOURCE_ITEM_SERVICE = 'ItemService'
    RESOURCE_TXN = 'Txn'


class QBXML_RESPONSE_STATUS_CODES:
    OK = '0'
    NO_MATCH = '1'
    ONE_OR_MORE_NOT_FOUND = '500'
    OBJECT_CAN_NOT_BE_RETURNED = '510'
    INTERNAL_ERROR = '1000'
    INVALID_OBJECT_ID = '3000'
    INVALID_BOOLEAN = '3010'
    INVALID_DATE = '3020'
    INVALID_DATE_RANGE = '3030'
    NAME_IS_NOT_UNIQUE = '3100'


def get_processors():
    processors = []
    for processor_class in qbwc_settings.RESPONSE_PROCESSORS:
        processors.append(processor_class)
    if not processors:
        raise ImproperlyConfigured(
            'No processors have been defined. Does '
            'RESPONSE_PROCESSORS for Quickbooks contain anything?'
        )
    return processors


def get_realm_model():
    return qbwc_settings.REALM_MODEL_CLASS


def get_realm_session_model():
    return qbwc_settings.REALM_SESSION_MODEL_CLASS


def get_qbd_task_model():
    return qbwc_settings.QBD_TASK_MODEL_CLASS



from django.http import HttpResponse
from django.views import View


class Support(View):

    def get(self, request, *args, **kwargs):
        html = "<html><body>Example of a supports page</body></html>"

        return HttpResponse(html)



from abc import abstractmethod
from uuid import uuid4, uuid1

from django.contrib.auth.hashers import make_password, check_password
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from lxml import etree

from django_quickbooks import qbwc_settings, QUICKBOOKS_ENUMS
from django_quickbooks.exceptions import QBOperationNotFound
from django_quickbooks.managers import RealmQuerySet, RealmSessionQuerySet, QBDTaskQuerySet
from django_quickbooks.objects import import_object_cls


class RealmMixin(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid4)
    name = models.CharField(max_length=155)
    password = models.CharField(max_length=128, null=True)

    objects = RealmQuerySet.as_manager()

    def set_password(self, raw_password):
        self.password = make_password(raw_password)
        self._password = raw_password

    def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """

        def setter(raw_password):
            self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            self.save(update_fields=["password"])

        return check_password(raw_password, self.password, setter)

    class Meta:
        abstract = True


class RealmSessionMixin(models.Model):
    id = models.UUIDField(verbose_name='Ticket for QBWC session', primary_key=True, editable=False, default=uuid1)
    created_at = models.DateTimeField(auto_now_add=True)
    ended_at = models.DateTimeField(null=True)

    objects = RealmSessionQuerySet.as_manager()

    class Meta:
        abstract = True


class QBDTaskMixin(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    qb_operation = models.CharField(max_length=25)
    qb_resource = models.CharField(max_length=50)
    object_id = models.CharField(max_length=50, null=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, null=True)
    content_object = GenericForeignKey()
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True

    objects = QBDTaskQuerySet.as_manager()

    def get_request(self):
        obj_class = import_object_cls(self.qb_resource)
        service = obj_class.get_service()()
        obj = self.content_object if self.content_type and self.object_id else None

        if self.qb_operation == QUICKBOOKS_ENUMS.OPP_QR:
            return service.all()
        elif not obj:
            raise ObjectDoesNotExist
        elif self.qb_operation == QUICKBOOKS_ENUMS.OPP_MOD:
            return service.update(obj.to_qbd_obj())
        elif self.qb_operation == QUICKBOOKS_ENUMS.OPP_ADD:
            return service.add(obj.to_qbd_obj())
        elif self.qb_operation == QUICKBOOKS_ENUMS.OPP_DEL:
            return service.delete(obj.to_qbd_obj())
        elif self.qb_operation == QUICKBOOKS_ENUMS.OPP_VOID:
            return service.void(obj.to_qbd_obj())
        else:
            raise QBOperationNotFound


# Below models are concrete implementations of above classes
# As initially I was working with django-tenant-schemas package I had to convert to that architecture
# Thus, below models are extended with schema_name concept that is the core of the django-tenant-schemas package:
# For more information: https://github.com/bernardopires/django-tenant-schemas

# NOTICE: you also find decorators in the package that only work with django-tenant-schemas


class Realm(RealmMixin):
    schema_name = models.CharField(max_length=100, unique=True)
    is_active = models.BooleanField(default=True)

    class Meta:
        abstract = False


class RealmSession(RealmSessionMixin):
    realm = models.ForeignKey(Realm, on_delete=models.CASCADE, related_name='sessions')

    class Meta:
        abstract = False


class QBDTask(QBDTaskMixin):
    realm = models.ForeignKey(Realm, on_delete=models.CASCADE, related_name='qb_tasks')

    class Meta:
        abstract = False


class QBDModelMixin(models.Model):
    qbd_object_id = models.CharField(max_length=127, unique=True, null=True, editable=False)
    qbd_object_updated_at = models.DateTimeField(null=True, editable=False)
    qbd_object_version = models.CharField(max_length=127, null=True, editable=False)

    class Meta:
        abstract = True

    @property
    def is_qbd_obj_created(self):
        return self.qbd_object_id and self.qbd_object_version

    @abstractmethod
    def to_qbd_obj(self, **fields):
        pass

    @classmethod
    def from_qbd_obj(cls, qbd_obj):
        return None


def create_qwc(realm, **kwargs):
    root = etree.Element("QBWCXML")

    app_name = kwargs.pop('app_name', qbwc_settings.APP_NAME)
    msg = etree.SubElement(root, 'AppName')
    msg.text = str(app_name)

    app_id = kwargs.pop('app_id', qbwc_settings.APP_ID)
    msg = etree.SubElement(root, 'AppID')
    msg.text = str(app_id)

    app_url = kwargs.pop('app_url', qbwc_settings.APP_URL)
    msg = etree.SubElement(root, 'AppURL')
    msg.text = str(app_url)

    app_desc = kwargs.pop('app_desc', qbwc_settings.APP_DESCRIPTION)
    msg = etree.SubElement(root, 'AppDescription')
    msg.text = str(app_desc)

    app_support = kwargs.pop('app_support', qbwc_settings.APP_SUPPORT)
    msg = etree.SubElement(root, 'AppSupport')
    msg.text = str(app_support)

    username = realm.id
    msg = etree.SubElement(root, 'UserName')
    msg.text = str(username)

    owner_id = kwargs.pop('owner_id', qbwc_settings.OWNER_ID)
    msg = etree.SubElement(root, 'OwnerID')
    msg.text = str(owner_id)

    file_id = kwargs.pop('file_id', '{%s}' % uuid1())
    msg = etree.SubElement(root, 'FileID')
    msg.text = str(file_id)

    qb_type = kwargs.pop('qb_type', qbwc_settings.QB_TYPE)
    msg = etree.SubElement(root, 'QBType')
    msg.text = str(qb_type)

    msg = etree.SubElement(root, 'Scheduler')

    schedule_n_minutes = kwargs.pop('schedule_n_minutes', qbwc_settings.MINIMUM_RUN_EVERY_NMINUTES)
    n_minutes = etree.SubElement(msg, 'RunEveryNMinutes')
    n_minutes.text = str(schedule_n_minutes)

    tree = etree.ElementTree(root)

    return etree.tostring(tree, xml_declaration=True, encoding='UTF-8')



import random
import string
from importlib import import_module

from django.utils.six import string_types


def import_callable(path_or_callable):
    if hasattr(path_or_callable, '__call__'):
        return path_or_callable
    else:
        assert isinstance(path_or_callable, string_types)
        package, attr = path_or_callable.rsplit('.', 1)
        return getattr(import_module(package), attr)


def xml_encode(value: str):
    transform = {
        '&': '&amp;',
        '<': '&lt;',
        '>': '&gt;',
        '"': '&quot;',
    }
    for transform_key, transform_value in transform.items():
        value = value.replace(transform_key, transform_value)
    return value


def xml_setter(name, value, encode=False, **options):
    option_xml = ''

    for option_key, option_value in options.items():
        option_value = xml_encode(option_value) if encode else option_value
        option_xml += '%s="%s" ' % (option_key, option_value)
    option_xml = option_xml.strip()

    value = xml_encode(value) if encode else value
    return '<%s>%s</%s>' % (name, value, name) if not option_xml else '<%s %s>%s</%s>' % (name, option_xml, value, name)


def get_xml_meta_info():
    return '<?xml version="1.0"?><?qbxml version="13.0"?>'


def random_string(length=10):
    letters = string.ascii_lowercase + string.ascii_uppercase
    return ''.join(random.choice(letters) for i in range(length))


def convert_qbd_model_to_qbdtask(obj, qb_resource, qb_operation=None, **kwargs):
    from django_quickbooks import QUICKBOOKS_ENUMS
    from django.contrib.contenttypes.models import ContentType

    if not qb_operation:
        if obj.is_qbd_obj_created:
            qb_operation = QUICKBOOKS_ENUMS.OPP_MOD
        else:
            qb_operation = QUICKBOOKS_ENUMS.OPP_ADD

        return dict(
            qb_operation=qb_operation,
            qb_resource=qb_resource,
            object_id=obj.id,
            content_type=ContentType.objects.get_for_model(obj),
        )



from spyne.decorator import rpc
from spyne.model.complex import Array, Unicode
from spyne.model.primitive import Integer, String
from spyne.service import ServiceBase

from django_quickbooks import QBWC_CODES, HIGHEST_SUPPORTING_QBWC_VERSION, \
    get_session_manager
from django_quickbooks.signals import realm_authenticated


class QuickBooksService(ServiceBase):

    @rpc(Unicode, Unicode, _returns=Array(Unicode))
    def authenticate(ctx, strUserName, strPassword):
        """
        Authenticate the web connector to access this service.

        @param strUserName user name to use for authentication
        @param strPassword password to use for authentication

        @return the completed array
        """
        print('authenticate()')
        return_array = []
        realm = session_manager.authenticate(username=strUserName, password=strPassword)
        if realm and realm.is_active:
            realm_authenticated.send(sender=realm.__class__, realm=realm)
            if not session_manager.in_session(realm):
                session_manager.add_new_jobs(realm)
                if session_manager.new_jobs(realm):
                    ticket = session_manager.set_ticket(realm)
                    return_array.append(ticket)
                    return_array.append(QBWC_CODES.CC)
                    # TODO: need to think about appropriate management of delays
                    # returnArray.append(str(settings.QBWC_UPDATE_PAUSE_SECONDS))
                    # returnArray.append(str(settings.QBWC_MINIMUM_UPDATE_SECONDS))
                    # returnArray.append(str(settings.QBWC_MINIMUM_RUN_EVERY_NSECONDS))
                else:
                    return_array.append(QBWC_CODES.NONE)
                    return_array.append(QBWC_CODES.NONE)
            else:
                return_array.append(QBWC_CODES.BUSY)
                return_array.append(QBWC_CODES.BUSY)
        else:
            return_array.append(QBWC_CODES.NVU)
            return_array.append(QBWC_CODES.NVU)

        print('authenticate(): authenticated_array=%s' % return_array)
        return return_array

    @rpc(Unicode, _returns=Unicode)
    def clientVersion(ctx, strVersion):
        """
        Evaluate the current web connector version and react to it.

        @param strVersion the version of the QB web connector

        @return string telling the web connector what to do next.
        """

        print('clientVersion(): version=%s' % strVersion)
        # TODO: add version checker for types: warning, error, ok
        return QBWC_CODES.CV

    @rpc(Unicode, _returns=Unicode)
    def closeConnection(ctx, ticket):
        """
        Tell the web service that the web connector is finished with the updated session.

        @param ticket the ticket from web connector supplied by web service during call to authenticate method
        @return string telling the web connector what to do next.
        """
        print('closeConnection(): ticket=%s' % ticket)
        session_manager.clear_ticket(ticket)
        return QBWC_CODES.CONN_CLS_OK

    @rpc(Unicode, Unicode, Unicode, _returns=Unicode)
    def connectionError(ctx, ticket, hresult, message):
        """
        Tell the web service about an error the web connector encountered in its attempt to connect to QuickBooks
        or QuickBooks POS

        @param ticket the ticket from web connector supplied by web service during call to authenticate method
        @param hresult the HRESULT (in HEX) from the exception thrown by the request processor
        @param message The error message that accompanies the HRESULT from the request processor

        @return string value "done" to indicate web service is finished or the full path of the different company for
        retrying _set_connection.
        """
        print('connectionError(): ticket=%s, hresult=%s, message=%s' % (ticket, hresult, message))
        session_manager.clear_ticket(ticket)
        return QBWC_CODES.CONN_CLS_ERR

    @rpc(Unicode, _returns=Unicode)
    def getLastError(ctx, ticket):
        """
        Allow the web service to return the last web service error, normally for displaying to user, before
        causing the update action to stop.

        @param ticket the ticket from web connector supplied by web service during call to authenticate method

        @return string message describing the problem and any other information that you want your user to see.
        The web connector writes this message to the web connector log for the user and also displays it in the web
        connector’s Status column.
        """
        print('getLastError(): ticket=%s' % ticket)
        return QBWC_CODES.UNEXP_ERR

    @rpc(Unicode, _returns=Unicode)
    def getServerVersion(ctx, ticket):
        """
        Provide a way for web-service to notify web connector of it’s version and other details about version

        @param ticket the ticket from web connector supplied by web service during call to authenticate method

        @return string message string describing the server version and any other information that user may want to see
        """
        print('getServerVersion(): version=%s' % HIGHEST_SUPPORTING_QBWC_VERSION)
        return HIGHEST_SUPPORTING_QBWC_VERSION

    @rpc(Unicode, _returns=Unicode)
    def interactiveDone(ctx, ticket):
        """
        Allow the web service to indicate to web connector that it is done with interactive mode.

        @param ticket the ticket from web connector supplied by web service during call to authenticate method

        @return string value "Done" should be returned when interactive session is over
        """
        print('interactiveDone(): ticket=%s' % ticket)
        return QBWC_CODES.INTR_DONE

    @rpc(Unicode, Unicode, _returns=Unicode)
    def interactiveRejected(ctx, ticket, reason):
        """
        Allow the web service to take alternative action when the interactive session it requested was
        rejected by the user or by timeout in the absence of the user.

        @param ticket the ticket from web connector supplied by web service during call to authenticate method
        @param reason the reason for the rejection of interactive mode

        @return string value "Done" should be returned when interactive session is over
        """
        print('interactiveRejected()')
        print(ticket)
        print(reason)
        return 'Interactive mode rejected'

    @rpc(Unicode, Unicode, Unicode, Unicode, _returns=Integer)
    def receiveResponseXML(ctx, ticket, response, hresult, message):
        """
        Return the data request response from QuickBooks or QuickBooks POS.

        @param ticket the ticket from web connector supplied by web service during call to authenticate method
        @param response qbXML response from QuickBooks or qbposXML response from QuickBooks POS
        @param hresult  The hresult and message could be returned as a result of certain errors that could occur when
        QuickBooks or QuickBooks POS sends requests is to the QuickBooks/QuickBooks POS request processor via the
        ProcessRequest call
        @param message The error message that accompanies the HRESULT from the request processor

        @return int a positive integer less than 100 represents the percentage of work completed. A value of 1 means one
        percent complete, a value of 100 means 100 percent complete--there is no more work. A negative value means an
        error has occurred and the Web Connector responds to this with a getLastError call.
        """
        print('receiveResponseXML()')
        print("ticket=" + ticket)
        print("response=" + response)
        if hresult:
            print("hresult=" + hresult)
            print("message=" + message)

        return session_manager.process_response(ticket, response, hresult, message)

    @rpc(Unicode, Unicode, Unicode, Unicode, Integer, Integer, _returns=String)
    def sendRequestXML(ctx, ticket, strHCPResponse, strCompanyFileName, qbXMLCountry, qbXMLMajorVers, qbXMLMinorVers):
        print('sendRequestXML() has been called')
        print('ticket:', ticket)
        print('strHCPResponse', strHCPResponse)
        print('strCompanyFileName', strCompanyFileName)
        print('qbXMLCountry', qbXMLCountry)

        return session_manager.get_request(ticket)

    @rpc(Unicode, Unicode, _returns=Unicode)
    def interactiveUrl(ctx, ticket, sessionID):
        print('interactiveUrl')
        print(ticket)
        print(sessionID)
        return ''


session_manager = get_session_manager()



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.objects.base import BaseObject
from django_quickbooks.validators import SchemeValidator


class ItemService(BaseObject):
    fields = dict(
        ListID=dict(validator=dict(type=SchemeValidator.IDTYPE)),
        Name=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        FullName=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Parent=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
    )

    @staticmethod
    def get_service():
        from django_quickbooks.services.item_service import ServiceOfItemService
        return ServiceOfItemService


class InvoiceLine(BaseObject):
    fields = dict(
        TxnLineID=dict(validator=dict(type=SchemeValidator.IDTYPE)),
        Item=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        Desc=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Quantity=dict(validator=dict(type=SchemeValidator.FLOATTYPE)),
        Rate=dict(validator=dict(type=SchemeValidator.FLOATTYPE)),
    )


class Invoice(BaseObject):
    fields = dict(
        TxnID=dict(validator=dict(type=SchemeValidator.IDTYPE)),
        TimeCreated=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        TimeModified=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        EditSequence=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        TxnDate=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Customer=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        BillAddress=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        ShipAddress=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        IsPending=dict(validator=dict(type=SchemeValidator.BOOLTYPE)),
        DueDate=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Memo=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        InvoiceLine=dict(many=True, validator=dict(type=SchemeValidator.OBJTYPE)),
    )

    @staticmethod
    def get_service():
        from django_quickbooks.services.invoice import InvoiceService
        return InvoiceService


class Txn(BaseObject):
    fields = dict(
        TxnID=dict(validator=dict(type=SchemeValidator.IDTYPE)),
        TxnType=dict(validator=dict(type=SchemeValidator.STRTYPE)),
    )

    def as_xml(self, class_name=None, indent=0, opp_type=QUICKBOOKS_ENUMS.OPP_ADD,
               version=QUICKBOOKS_ENUMS.VERSION_13, **kwargs):
        xml = super(Txn, self).as_xml(class_name, indent, opp_type, version,  **kwargs)

        return xml\
            .replace(f'<{__class__.__name__}{opp_type}>', '')\
            .replace(f'</{__class__.__name__}{opp_type}>', '')\
            .replace('<TxnType>', f'<Txn{opp_type}Type>')\
            .replace('</TxnType>', f'</Txn{opp_type}Type>')



from django_quickbooks.queue_manager import RabbitMQManager


class RequestBuilder(RabbitMQManager):

    def process(self, realm_id, request_body):
        self.publish_message(request_body, realm_id)
        self._get_channel().close()
        self._get_connection().close()





from django_quickbooks.processors.base import \
    ResponseProcessor

from django_quickbooks.processors.customer import \
    CustomerQueryResponseProcessor, \
    CustomerAddResponseProcessor, \
    CustomerModResponseProcessor

from django_quickbooks.processors.invoice import \
    InvoiceQueryResponseProcessor, \
    InvoiceAddResponseProcessor, \
    InvoiceModResponseProcessor

from django_quickbooks.processors.item_service import \
    ItemServiceQueryResponseProcessor



from django_quickbooks.objects.base import BaseObject
from django_quickbooks.validators import SchemeValidator


class Address(BaseObject):
    fields = dict(
        Addr1=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Addr2=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Addr3=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Addr4=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Addr5=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        City=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        State=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        PostalCode=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Country=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Note=dict(validator=dict(type=SchemeValidator.STRTYPE)),
    )


class BillAddress(Address):
    pass


class ShipAddress(Address):
    pass



import pytest
from lxml import etree

from django_quickbooks.exceptions import ValidationError
from django_quickbooks.objects import Customer, BillAddress, ShipAddress, Invoice, InvoiceLine, ItemService


def test_create_invoice_with_validation_error():
    with pytest.raises(ValidationError):
        Invoice(
            TxnID=12345,
            TimeCreated=12345,
            TimeModified=12345,
            EditSequence=12345,
            TxnDate=12345,
            IsPending=12345,
            DueDate=12345,
            Memo=12345,
            AltPhone=1234,
            Fax=1234,
            Email=1234,
            Contact=1234,
            AltContact=1234
        )


def test_create_invoice(
        sample_address_data,
        sample_customer_data,
        sample_invoice_item_service1_data,
        sample_invoice_item_service2_data,
        sample_invoice_line1_data,
        sample_invoice_line2_data,
        sample_invoice_data,
):
    invoice = Invoice(
        **sample_invoice_data,
        BillAddress=BillAddress(**sample_address_data),
        ShipAddress=ShipAddress(**sample_address_data),
        Customer=Customer(**sample_customer_data),
        InvoiceLine=[
            InvoiceLine(**sample_invoice_line1_data, Item=ItemService(**sample_invoice_item_service1_data)),
            InvoiceLine(**sample_invoice_item_service2_data, Item=ItemService(**sample_invoice_item_service2_data))
        ]
    )

    assert invoice.TxnID == sample_invoice_data['TxnID']
    assert invoice.TimeCreated == sample_invoice_data['TimeCreated']
    assert invoice.TimeModified == sample_invoice_data['TimeModified']
    assert invoice.EditSequence == sample_invoice_data['EditSequence']
    assert invoice.TxnDate == sample_invoice_data['TxnDate']
    assert invoice.IsPending == sample_invoice_data['IsPending']
    assert invoice.DueDate == sample_invoice_data['DueDate']
    assert invoice.Memo == sample_invoice_data['Memo']

    assert isinstance(invoice.BillAddress, BillAddress)
    assert isinstance(invoice.ShipAddress, ShipAddress)
    assert isinstance(invoice.Customer, Customer)
    assert isinstance(invoice.InvoiceLine, list)
    assert isinstance(invoice.InvoiceLine[0], InvoiceLine)
    assert isinstance(invoice.InvoiceLine[0].Item, ItemService)
    assert isinstance(invoice.InvoiceLine[1], InvoiceLine)


def test_create_invoice_from_xml(
        sample_invoice_xml,
        sample_address_data,
        sample_invoice_customer_data,
        sample_invoice_item_service1_data,
        sample_invoice_item_service2_data,
        sample_invoice_line1_data,
        sample_invoice_line2_data,
        sample_invoice_data,

):
    root_lxml = etree.fromstring(sample_invoice_xml)
    invoice = Invoice.from_lxml(root_lxml)

    assert invoice.TxnID == sample_invoice_data['TxnID']
    assert invoice.TimeCreated == sample_invoice_data['TimeCreated']
    assert invoice.TimeModified == sample_invoice_data['TimeModified']
    assert invoice.EditSequence == sample_invoice_data['EditSequence']
    assert invoice.IsPending == sample_invoice_data['IsPending']
    assert invoice.DueDate == sample_invoice_data['DueDate']
    assert invoice.Memo == sample_invoice_data['Memo']

    assert invoice.BillAddress == BillAddress(**sample_address_data)
    assert invoice.ShipAddress == ShipAddress(**sample_address_data)
    assert invoice.Customer == Customer(**sample_invoice_customer_data)

    assert isinstance(invoice.InvoiceLine, list)
    assert invoice.InvoiceLine[0] == InvoiceLine(
        **sample_invoice_line1_data,
        Item=ItemService(**sample_invoice_item_service1_data)
    )
    assert invoice.InvoiceLine[1] == InvoiceLine(
        **sample_invoice_line2_data,
        Item=ItemService(**sample_invoice_item_service2_data)
    )



from django.utils.translation import ugettext_lazy as _


class ValidationCode:
    REQUIRED = 'required'
    MIN_LENGTH = 'min_length'
    MAX_LENGTH = 'max_length'
    INVALID_TYPE = 'invalid_type'


VALIDATION_MESSAGES = {
    ValidationCode.REQUIRED: _('This field is required'),
    ValidationCode.MIN_LENGTH: _('The minimum length for the field is %s'),
    ValidationCode.MAX_LENGTH: _('The maximum length for the field is %s'),
    ValidationCode.INVALID_TYPE: _('Invalid type %s for the field type %s'),
}


def _get_error_details(detail, code):
    return {code: detail}


class QbException(Exception):
    detail = 'Quickbooks exception error'

    def __init__(self, error=None):
        if error:
            self.detail = error


class ValidationError(Exception):
    default_detail = _('Invalid attribute.')
    default_code = 'invalid'

    def __init__(self, detail=None, code=None):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code

        # Several errors may be collected together, thus
        if not isinstance(detail, dict) and not isinstance(detail, list):
            detail = [detail]

        self.detail = _get_error_details(detail, code)


class ValidationOptionNotFound(QbException):
    pass


class QBXMLParseError(QbException):
    pass


class QBXMLStatusError(QbException):
    pass


class QBOperationNotFound(QbException):
    pass


class QBObjectNotImplemented(QbException):
    pass


class QBObjectServiceNotImplemented(QbException):
    pass



import logging
from abc import ABC

from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.exceptions import QBObjectServiceNotImplemented
from django_quickbooks.exceptions import ValidationError
from django_quickbooks.objects import import_object_cls
from django_quickbooks.utils import xml_setter
from django_quickbooks.validators import SchemeValidator, is_primitive, is_list


def _default_value_setter(key, type):
    if type == 'LISTTYPE':
        return []
    return None


class BaseObject(ABC):
    fields = dict()
    validator = SchemeValidator()

    def __setattr__(self, key, value):
        if key in self.fields:
            super().__setattr__(key, value)

    def __init__(self, **kwargs):
        errors = []
        # FIXME: ValidationError handling still seems clumsy, need to reviewed
        for field_name, value in kwargs.items():
            try:
                self.validator.validate(field_name, value, **self.fields[field_name])
                setattr(self, field_name, value)
            except ValidationError as exc:
                errors.append(exc.detail)

        if errors:
            logger = logging.getLogger('django.request')
            logging.error(errors)
            raise ValidationError(errors)

        for field_name, options in self.fields.items():
            if not hasattr(self, field_name):
                setattr(self,
                        field_name,
                        _default_value_setter(field_name, self.fields[field_name]['validator']['type'])
                        )

    def as_xml(self, class_name=None, indent=0, opp_type=QUICKBOOKS_ENUMS.OPP_ADD,
               version=QUICKBOOKS_ENUMS.VERSION_13, **kwargs):
        xml = ''
        for field_key, options in self.fields.items():
            if hasattr(self, field_key):
                obj = getattr(self, field_key)
                if not obj:
                    continue
                if is_primitive(obj):
                    if self.fields[field_key]['validator']['type'] == self.validator.BOOLTYPE:
                        xml += xml_setter(field_key, str(obj).lower(), encode=True)
                    else:
                        xml += xml_setter(field_key, str(obj), encode=True)

                elif is_list(obj):
                    for element in obj:
                        if is_primitive(element):
                            if self.fields[field_key]['validator']['type'] == self.validator.BOOLTYPE:
                                xml += xml_setter(field_key, str(element).lower(), encode=True)
                            else:
                                xml += xml_setter(field_key, str(element), encode=True)
                        else:
                            if field_key in kwargs.get('ref_fields', []):
                                xml += element.as_xml(class_name=field_key, indent=indent,
                                                      opp_type=QUICKBOOKS_ENUMS.OBJ_REF, version=version, **kwargs)
                            elif field_key in kwargs.get('change_fields', []):
                                xml += element.as_xml(class_name=field_key, indent=indent, opp_type=opp_type,
                                                      version=version, **kwargs)
                            elif field_key in kwargs.get('complex_fields', []):
                                xml += element.as_xml(class_name=field_key, indent=indent, opp_type='', version=version,
                                                      **kwargs)
                else:
                    if field_key in kwargs.get('ref_fields', []):
                        xml += obj.as_xml(class_name=field_key, indent=indent, opp_type=QUICKBOOKS_ENUMS.OBJ_REF,
                                          version=version, **kwargs)
                    elif field_key in kwargs.get('change_fields', []):
                        xml += obj.as_xml(class_name=field_key, indent=indent, opp_type=opp_type, version=version,
                                          **kwargs)
                    elif field_key in kwargs.get('complex_fields', []):
                        xml += obj.as_xml(class_name=field_key, indent=indent, opp_type='', version=version, **kwargs)

        class_name = self.__class__.__name__ if not class_name else class_name
        xml = xml_setter(class_name + opp_type, xml)
        return xml

    @classmethod
    def from_lxml(cls, lxml_obj):

        def to_internal_value(field, type):
            converters = dict(
                STRTYPE=lambda x: x.text,
                ESTYPE=lambda x: x.text,
                IDTYPE=lambda x: x.text,
                FLOATTYPE=lambda x: float(x.text),
                BOOLTYPE=lambda x: bool(x.text),
                OBJTYPE=lambda x: import_object_cls(x.tag).from_lxml(x),
            )
            return converters[type](field) if type in converters else None

        obj_data = dict()
        for field in list(lxml_obj):
            field_name = field.tag
            value = None
            if field_name in cls.fields:
                value = to_internal_value(field, cls.fields[field_name]['validator']['type'])
            elif field_name == 'ParentRef':
                field_name = 'Parent'
                value = cls.from_lxml(field)
            elif isinstance(field_name, str) and any(
                    index == len(field_name) - 3 for index in list(map(field_name.find, ['Ref', 'Ret']))):
                field_name = field_name[:len(field_name) - 3]
                if field_name in cls.fields:
                    field.tag = field_name
                    value = to_internal_value(field, cls.fields[field_name]['validator']['type'])
            if value:
                if field_name in cls.fields and cls.fields[field_name].get('many', False):
                    if field_name in obj_data:
                        obj_data[field_name].append(value)
                    else:
                        obj_data[field_name] = [value]
                else:
                    obj_data[field_name] = value

        return cls(**obj_data)

    def __eq__(self, other):
        for field_name, options in self.fields.items():
            if not hasattr(self, field_name) and hasattr(other, field_name):
                return False
            if not hasattr(other, field_name) and hasattr(self, field_name):
                return False
            if getattr(self, field_name) != getattr(other, field_name):
                return False
        return True

    @staticmethod
    def get_service():
        raise QBObjectServiceNotImplemented



from copy import deepcopy

import pytest

from django_quickbooks.objects import Invoice, BillAddress, ShipAddress, Customer, InvoiceLine, ItemService
from django_quickbooks.services.invoice import InvoiceService


@pytest.fixture
def invoice_service():
    return InvoiceService()


@pytest.fixture
def sample_invoice_add(
        sample_invoice_data,
        sample_address_data,
        sample_customer_data,
        sample_invoice_item_service1_data,
        sample_invoice_item_service2_data,
        sample_invoice_line1_data,
        sample_invoice_line2_data
):
    invoice_data = deepcopy(sample_invoice_data)
    invoice_data.pop('TxnID')
    invoice_data.pop('EditSequence')
    invoice_data.pop('TimeCreated')
    invoice_data.pop('TimeModified')
    invoice_line1_data = deepcopy(sample_invoice_line1_data)
    invoice_line2_data = deepcopy(sample_invoice_line2_data)
    invoice_line1_data.pop('TxnLineID')
    invoice_line2_data.pop('TxnLineID')
    sample_invoice_item_service1_data.pop('FullName')
    sample_invoice_item_service2_data.pop('FullName')
    customer_data = dict(ListID=sample_customer_data['ListID'], FullName=sample_customer_data['FullName'])

    return Invoice(
        **invoice_data,
        BillAddress=BillAddress(**sample_address_data),
        ShipAddress=ShipAddress(**sample_address_data),
        Customer=Customer(**customer_data),
        InvoiceLine=[
            InvoiceLine(**invoice_line1_data, Item=ItemService(**sample_invoice_item_service1_data)),
            InvoiceLine(**invoice_line2_data, Item=ItemService(**sample_invoice_item_service2_data))
        ]
    )


@pytest.fixture
def sample_invoice_add_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <InvoiceAddRq>
            <InvoiceAdd>
                <TxnDate>2018-03-24T00:31:04+05:00</TxnDate>
                <CustomerRef>
                    <ListID>800004ED-1525972764</ListID>
                    <FullName>Amazon</FullName>
                </CustomerRef>
                <BillAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </BillAddress>
                <IsPending>true</IsPending>
                <DueDate>2018-03-24T00:31:04+05:00</DueDate>
                <Memo>a2e8929f-6f03-4bc2-ad27-ea6aa2cc11cd</Memo>
                <InvoiceLineAdd>
                    <ItemRef>
                        <ListID>800004ED-1525972764</ListID>
                    </ItemRef>
                    <Desc>First Line of Invoice</Desc>
                    <Quantity>1.0</Quantity>
                    <Rate>500.5</Rate>
                </InvoiceLineAdd>
                <InvoiceLineAdd>
                    <ItemRef>
                        <ListID>800004ED-1525972765</ListID>
                    </ItemRef>
                    <Desc>Second Line of Invoice</Desc>
                    <Quantity>1.0</Quantity>
                    <Rate>100.5</Rate>
                </InvoiceLineAdd>
            </InvoiceAdd>
        </InvoiceAddRq>
    </QBXMLMsgsRq>
</QBXML>
    """


def test_invoice_service_add(
        invoice_service,
        sample_invoice_add,
        sample_invoice_add_xml,
        xml_compare_fn
):
    checking_xml = invoice_service.add(sample_invoice_add)
    xml_compare_fn(checking_xml, sample_invoice_add_xml)


@pytest.fixture
def sample_invoice_mod(
        sample_invoice_add,
        sample_invoice_line1_data,
        sample_invoice_line2_data,
        sample_invoice_data
):
    invoice = sample_invoice_add
    invoice.TxnID = sample_invoice_data['TxnID']
    invoice.EditSequence = sample_invoice_data['EditSequence']
    invoice.InvoiceLine[0].TxnLineID = sample_invoice_line1_data['TxnLineID']
    invoice.InvoiceLine[1].TxnLineID = sample_invoice_line2_data['TxnLineID']

    return invoice


@pytest.fixture
def sample_invoice_mod_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <InvoiceModRq>
            <InvoiceMod>
                <TxnID>800004ED-1525972764</TxnID>
                <EditSequence>1525972764</EditSequence>
                <TxnDate>2018-03-24T00:31:04+05:00</TxnDate>
                <CustomerRef>
                    <ListID>800004ED-1525972764</ListID>
                    <FullName>Amazon</FullName>
                </CustomerRef>
                <BillAddress>
                    <Addr1>2305 Litton Ln</Addr1>
                    <City>Hebron</City>
                    <State>Kentucky</State>
                    <PostalCode>41048</PostalCode>
                    <Country>United States</Country>
                    <Note>Nice address</Note>
                </BillAddress>
                <IsPending>true</IsPending>
                <DueDate>2018-03-24T00:31:04+05:00</DueDate>
                <Memo>a2e8929f-6f03-4bc2-ad27-ea6aa2cc11cd</Memo>
                <InvoiceLineMod>
                    <TxnLineID>800004ED-1525972764</TxnLineID>
                    <ItemRef>
                        <ListID>800004ED-1525972764</ListID>
                    </ItemRef>
                    <Desc>First Line of Invoice</Desc>
                    <Quantity>1.0</Quantity>
                    <Rate>500.5</Rate>
                </InvoiceLineMod>
                <InvoiceLineMod>
                    <TxnLineID>800004ED-1525972765</TxnLineID>
                    <ItemRef>
                        <ListID>800004ED-1525972765</ListID>
                    </ItemRef>
                    <Desc>Second Line of Invoice</Desc>
                    <Quantity>1.0</Quantity>
                    <Rate>100.5</Rate>
                </InvoiceLineMod>
            </InvoiceMod>
        </InvoiceModRq>
    </QBXMLMsgsRq>
</QBXML>
    """


def test_invoice_service_update(
        invoice_service,
        sample_invoice_mod,
        sample_invoice_mod_xml,
        xml_compare_fn
):
    checking_xml = invoice_service.update(sample_invoice_mod)
    xml_compare_fn(checking_xml, sample_invoice_mod_xml)


@pytest.fixture
def sample_invoice_query_all_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
        <QBXML>
            <QBXMLMsgsRq onError="stopOnError">
                <InvoiceQueryRq metaData="NoMetaData" iterator="Start">
                    <MaxReturned>100</MaxReturned>
                </InvoiceQueryRq>
            </QBXMLMsgsRq>
        </QBXML>
    """


def test_invoice_service_query_all(
        invoice_service,
        sample_invoice_query_all_xml,
        xml_compare_fn
):
    checking_xml = invoice_service.all()
    xml_compare_fn(checking_xml, sample_invoice_query_all_xml)


@pytest.fixture
def sample_invoice_query_id_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <InvoiceQueryRq metaData="NoMetaData">
            <TxnID>800004ED-1525972764</TxnID>
        </InvoiceQueryRq>
    </QBXMLMsgsRq>
</QBXML>
    """


def test_invoice_service_query_id(invoice_service, sample_invoice_query_id_xml, xml_compare_fn):
    checking_xml = invoice_service.find_by_id('800004ED-1525972764')
    xml_compare_fn(checking_xml, sample_invoice_query_id_xml)


@pytest.fixture
def sample_invoice_void_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <TxnVoidRq>
            <TxnID>800004ED-1525972764</TxnID>
            <TxnVoidType>Invoice</TxnVoidType>
        </TxnVoidRq>
    </QBXMLMsgsRq>
</QBXML>
    """


def test_invoice_service_void(invoice_service, sample_invoice_mod, sample_invoice_void_xml, xml_compare_fn):
    checking_xml = invoice_service.void(sample_invoice_mod)
    xml_compare_fn(checking_xml, sample_invoice_void_xml)


@pytest.fixture
def sample_invoice_delete_xml():
    return """<?xml version="1.0"?><?qbxml version="13.0"?>
<QBXML>
    <QBXMLMsgsRq onError="stopOnError">
        <TxnDelRq>
            <TxnID>800004ED-1525972764</TxnID>
            <TxnDelType>Invoice</TxnDelType>
        </TxnDelRq>
    </QBXMLMsgsRq>
</QBXML>
    """


def test_invoice_service_delete(invoice_service, sample_invoice_mod, sample_invoice_delete_xml, xml_compare_fn):
    checking_xml = invoice_service.delete(sample_invoice_mod)
    xml_compare_fn(checking_xml, sample_invoice_delete_xml)



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.utils import get_xml_meta_info, xml_setter


class Service:
    ref_fields = []
    add_fields = []
    complex_fields = []
    mod_fields = []

    def _prepare_request(self, req_body):
        xml = xml_setter('QBXMLMsgsRq', req_body, onError='stopOnError')
        xml = xml_setter('QBXML', xml)
        xml = get_xml_meta_info() + xml
        return xml

    def _add(self, resource, object):
        xml = ''
        xml += xml_setter(resource + QUICKBOOKS_ENUMS.OPP_ADD + 'Rq', object.as_xml(
            opp_type=QUICKBOOKS_ENUMS.OPP_ADD, ref_fields=self.ref_fields, change_fields=self.add_fields, complex_fields=self.complex_fields))
        return self._prepare_request(xml)

    def _update(self, resource, object):
        xml = ''
        xml += xml_setter(resource + QUICKBOOKS_ENUMS.OPP_MOD + 'Rq', object.as_xml(
            opp_type=QUICKBOOKS_ENUMS.OPP_MOD, ref_fields=self.ref_fields, change_fields=self.mod_fields, complex_fields=self.complex_fields))
        return self._prepare_request(xml)

    def _void(self, resource, object):
        xml = ''
        xml += xml_setter(resource + QUICKBOOKS_ENUMS.OPP_VOID + 'Rq', object.as_xml(opp_type=QUICKBOOKS_ENUMS.OPP_VOID))
        return self._prepare_request(xml)

    def _delete(self, resource, object):
        xml = ''
        xml += xml_setter(resource + QUICKBOOKS_ENUMS.OPP_DEL + 'Rq', object.as_xml(opp_type=QUICKBOOKS_ENUMS.OPP_DEL))
        return self._prepare_request(xml)

    def _all(self, resource):
        xml = xml_setter('MaxReturned', 100)
        xml = xml_setter(resource + QUICKBOOKS_ENUMS.OPP_QR + 'Rq', xml, metaData='NoMetaData', iterator='Start')
        return self._prepare_request(xml)

    def _find_by_id(self, resource, id, field_name='ListID'):
        xml = xml_setter(field_name, id)
        xml = xml_setter(resource + QUICKBOOKS_ENUMS.OPP_QR + 'Rq', xml, metaData='NoMetaData')
        return self._prepare_request(xml)

    def _find_by_full_name(self, resource, full_name):
        xml = xml_setter('FullName', full_name)
        xml = xml_setter(resource + QUICKBOOKS_ENUMS.OPP_QR + 'Rq', xml, metaData='NoMetaData')
        return self._prepare_request(xml)



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.services.base import Service


class ServiceOfItemService(Service):
    def all(self):
        return self._all(QUICKBOOKS_ENUMS.RESOURCE_ITEM_SERVICE)

    def find_by_id(self, id):
        return self._find_by_id(QUICKBOOKS_ENUMS.RESOURCE_ITEM_SERVICE, id)

    def find_by_full_name(self, full_name):
        return self._find_by_full_name(QUICKBOOKS_ENUMS.RESOURCE_ITEM_SERVICE, full_name)



# Generated by Django 2.2.9 on 2020-01-22 11:40

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('django_quickbooks', '0003_qbtask'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='QBTask',
            new_name='QBDTask',
        ),
    ]



from django.utils.module_loading import import_string

from django_quickbooks.exceptions import QBObjectNotImplemented

ALTERNATIVES = {
    'ItemService': ('Item',),
}


def import_object_cls(resource_name):
    for class_name, alternatives in ALTERNATIVES.items():
        if resource_name in alternatives:
            resource_name = class_name

    try:
        return import_string('%s.%s' % (__package__, resource_name))
    except ImportError:
        raise QBObjectNotImplemented


from django_quickbooks.objects.customer import \
    Customer

from django_quickbooks.objects.address import \
    BillAddress, \
    ShipAddress

from django_quickbooks.objects.invoice import \
    ItemService, \
    InvoiceLine, \
    Invoice



from django_quickbooks.objects.base import BaseObject
from django_quickbooks.validators import SchemeValidator


class Customer(BaseObject):
    fields = dict(
        ListID=dict(validator=dict(type=SchemeValidator.IDTYPE)),
        TimeCreated=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        TimeModified=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        EditSequence=dict(validator=dict(type=SchemeValidator.ESTYPE)),
        Name=dict(required=False, validator=dict(type=SchemeValidator.STRTYPE)),
        FullName=dict(required=True, validator=dict(type=SchemeValidator.STRTYPE)),
        IsActive=dict(validator=dict(type=SchemeValidator.BOOLTYPE)),
        CompanyName=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        BillAddress=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        ShipAddress=dict(validator=dict(type=SchemeValidator.OBJTYPE)),
        Phone=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        AltPhone=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Fax=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Email=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        Contact=dict(validator=dict(type=SchemeValidator.STRTYPE)),
        AltContact=dict(validator=dict(type=SchemeValidator.STRTYPE)),
    )

    def __init__(self, Name=None, IsActive=None, **kwargs):
        if Name:
            self.Name = Name

        if IsActive:
            self.IsActive = IsActive

        super().__init__(**kwargs)

    @staticmethod
    def get_service():
        from django_quickbooks.services.customer import CustomerService
        return CustomerService



from django_quickbooks import QUICKBOOKS_ENUMS
from django_quickbooks.exceptions import VALIDATION_MESSAGES, ValidationCode
from django_quickbooks.exceptions import ValidationOptionNotFound, ValidationError


def obj_type_validator(value):
    from django_quickbooks.objects.base import BaseObject
    return isinstance(value, BaseObject)


def is_primitive(value):
    from django_quickbooks.objects.base import BaseObject
    return isinstance(type(value), type) and not isinstance(value, (BaseObject, list))


def operation_type(value):
    for key, val in value.fields.items():
        if val['validator']['type'] == SchemeValidator.IDTYPE:
            has_attr = hasattr(value, key)
            if 'Txn' in key and has_attr:
                return QUICKBOOKS_ENUMS.OPP_MOD
            if 'Txn' in key and not has_attr:
                return QUICKBOOKS_ENUMS.OPP_ADD
            if not has_attr:
                return QUICKBOOKS_ENUMS.OBJ_REF
            if has_attr:
                return QUICKBOOKS_ENUMS.OPP_ADD
    return ''


def is_list(value):
    if not isinstance(value, list):
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), str),
                              ValidationCode.INVALID_TYPE)


def str_type_validator(value):
    if not isinstance(value, str):
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), str),
                              ValidationCode.INVALID_TYPE)


def es_type_validator(value):
    if not isinstance(value, str) or not value.isnumeric():
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), str),
                              ValidationCode.INVALID_TYPE)


def id_type_validator(value):
    if not isinstance(value, str):
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), str),
                              ValidationCode.INVALID_TYPE)


def bool_type_validator(value):
    if value not in [1, 0, 'true', 'false', '1', '0']:
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), bool),
                              ValidationCode.INVALID_TYPE)


def min_length_validator(value, length):
    if len(value) < length:
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.MIN_LENGTH] % length, ValidationCode.MIN_LENGTH)


def max_length_validator(value, length):
    if len(value) > length:
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.MAX_LENGTH] % length, ValidationCode.MAX_LENGTH)


def float_type_validator(value):
    if not isinstance(value, float):
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), float),
                              ValidationCode.INVALID_TYPE)


def required_validator(value, required=False):
    if not value and required:
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.REQUIRED], ValidationCode.REQUIRED)


def many_validator(value, many=False):
    if not isinstance(value, list) and many:
        raise ValidationError(VALIDATION_MESSAGES[ValidationCode.INVALID_TYPE] % (type(value), list),
                              ValidationCode.INVALID_TYPE)


class SchemeValidator:
    STRTYPE = 'STRTYPE'
    ESTYPE = 'ESTYPE'
    IDTYPE = 'IDTYPE'
    BOOLTYPE = 'BOOLTYPE'
    OBJTYPE = 'OBJTYPE'
    FLOATTYPE = 'FLOATTYPE'

    type_validators = dict(
        STRTYPE=str_type_validator,
        ESTYPE=es_type_validator,
        IDTYPE=id_type_validator,
        BOOLTYPE=bool_type_validator,
        OBJTYPE=obj_type_validator,
        FLOATTYPE=float_type_validator,
    )
    option_validators = dict(
        min_length=min_length_validator,
        max_length=max_length_validator,
    )

    def validate(self, field_name, value, **options):
        errors = []

        required = options.pop('required', False)

        try:
            required_validator(value, required)
        except ValidationError as exc:
            errors.append(exc.detail)

        many = options.pop('many', False)

        try:
            # should be given list type but given something else
            many_validator(value, many)
        except ValidationError as exc:
            errors.append(exc.detail)

        if many:
            for single_value in value:
                try:
                    self.validate(single_value, **options)
                except ValidationError as exc:
                    errors.append(exc.detail)

            if errors:
                raise ValidationError(errors, field_name)

        validator = options.pop('validator')
        typ = validator['type']

        try:
            self.type_validators[typ](value)
        except ValidationError as exc:
            errors.append(exc.detail)

        for value, key in options.items():
            if value not in self.option_validators:
                raise ValidationOptionNotFound
            try:
                self.option_validators[value](key)
            except ValidationError as exc:
                errors.append(exc.detail)

        if errors:
            raise ValidationError(errors, field_name)
